# -*- coding: utf-8; -*-
#
# Copyright (c) 2016 Álan Crístoffer
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import numpy as np
import scipy.signal

import json


class ModelSimulation(object):
    def __init__(self, data):
        self.data = data
        self.model = None
        self.x0 = None
        self.U = None
        self.T = None

    def run(self):
        try:
            self.model = [np.array(x) for x in eval(self.data['model'])]
        except:
            return json.dumps({'error': 'Invalid Model'})

        try:
            self.x0 = np.array(eval(self.data['x0']))
        except:
            return json.dumps({'error': 'Invalid x0'})

        try:
            self.U = eval(self.data['u'])
        except:
            return json.dumps({'error': 'Invalid U'})

        try:
            self.T = eval(self.data['duration'])
        except:
            return json.dumps({'error': 'Invalid Duration'})

        try:
            tf = len(self.model) < 4

            if len(self.model) % 2 == 1:
                dt = float(self.model[-1])
            else:
                if len(self.model) == 2:
                    self.model = scipy.signal.tf2ss(*self.model)
                vals = scipy.linalg.eigvals(self.model[0])
                dt = max(0.1, float('%.1f' % (max(abs(np.real(vals))) / 5)))
                if type(self.U) == list:
                    dt = self.T / len(self.U)
                self.model = scipy.signal.cont2discrete(self.model, dt)

            if type(self.U) == list:
                self.T = list(range(len(self.U)))
            else:
                self.T = list(range(int(self.T // dt)))
                self.U = [self.U for _ in self.T]

            A, B, C, D, _ = self.model
            x = self.x0 if not tf else np.zeros((A.shape[0], 1))
            outputs = {'u': self.U, 't': self.T}

            if C.shape[0] == 1:
                outputs['y'] = []
            else:
                for i in range(C.shape[0]):
                    outputs['y%d' % (i + 1)] = []

            if not tf:
                for i in range(len(x.flatten())):
                    outputs['x%d' % (i + 1)] = []

            for k in self.T:
                x = A @ x + B * self.U[k]
                y = C @ x + D * self.U[k]

                if not tf:
                    for i in range(len(x.flatten())):
                        k = i + 1
                        outputs['x%d' % k].append(np.asscalar(x.flatten()[i]))

                if C.shape[0] > 1:
                    for i in range(C.shape[0]):
                        k = i + 1
                        outputs['y%d' % k].append(np.asscalar(y.flatten()[i]))
                else:
                    outputs['y'].append(np.asscalar(y))

            outputs['t'] = [dt * k for k in outputs['t']]
            return json.dumps(outputs)
        except Exception as e:
            return json.dumps({'error': f'Error in simulation ({str(e)})'})



# -*- coding: utf-8; -*-
#
# Copyright (c) 2016 Álan Crístoffer
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import errno
import inspect
import os
import socket
import time

import appdirs

import moirai

LOG_DIR = appdirs.user_log_dir(
    appname=moirai.__name__,
    appauthor=moirai.__author__,
    version=moirai.__version__)
os.makedirs(LOG_DIR, exist_ok=True)
LOG_FILE = open(os.path.join(LOG_DIR, 'moirai.log'), 'at')


def log_file_path():
    """
    Returns the path of the log file.
    """
    return os.path.join(LOG_DIR, 'moirai.log')


def decorate_all_methods(decorator):
    """
    Decorates every member of a class.
    """

    def decorate(cls):
        """
        Decorates every member of a class.
        """
        for name, func in inspect.getmembers(cls, inspect.isfunction):
            setattr(cls, name, decorator(func))
        return cls

    return decorate


def ignore_eagain(func):
    """
    errno.EAGAIN is generated in some platforms when the socket is
    non-blocking. This decorator ignores this exception, as it's not an error.
    """

    def decorate(ret, *kargs):
        """
        Decorator.
        """
        try:
            return func(*kargs)
        except socket.error as error:
            if error.errno == errno.EAGAIN:
                return ret
            else:
                raise error

    return decorate


def log_msg(msg):
    """
    Logs `msg` to log file.
    """
    msg = '%s: %s\n' % (time.ctime(), msg)
    LOG_FILE.write(msg)
    LOG_FILE.flush()


def log(func):
    """
    Logs every call to a class method.
    """

    def decorate(*kargs, **kwargs):
        """
        Logs every call to a class method.
        """
        func_name = []
        if func.__module__ != '__main__':
            func_name.append(func.__module__)
        func_name.append(func.__qualname__)
        func_name = '.'.join(func_name)

        args = []
        if kargs:
            args = [str(arg) for arg in kargs]
        if kwargs:
            args += ['%s=%s' % i for i in kwargs.items()]
        args = ', '.join(args)

        log_str = '%s(%s)' % (func_name, args)
        log_msg(log_str)
        return func(*kargs, **kwargs)

    return decorate


def dont_raise(func):
    """
    Decorate every member of a class to stop raising exceptions.
    """

    def decorate(*kargs, **kwargs):
        """
        Decorate every member of a class to stop raising exceptions.
        """
        try:
            return func(*kargs, **kwargs)
        except Exception as error:
            log_msg('An Exception occurred: %s' % error)
            print('An Exception occurred: %s' % error)

    return decorate



# -*- coding: utf-8; -*-
#
# Copyright (c) 2016 Álan Crístoffer
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import ahio
import math
import numpy
import os

from moirai.database import DatabaseV1

# Port types:
# export enum Types {
#     Digital = 1,
#     Analog = 2,
#     Input = 4,
#     Output = 8,
#     PWM = 16
# }


class ConfiguredHardware(object):
    def __init__(self):
        self.db = DatabaseV1()
        config = self.db.get_setting('hardware_configuration')

        if config is None:
            raise Exception('No hardware configured')

        ahio.clear_path()
        try:
            paths = os.environ['AHIO_PATH'].split(os.pathsep)
            for path in paths:
                if os.path.exists(path):
                    ahio.add_path(os.path.expanduser(path))
        except Exception:
            pass

        self.driver = ahio.new_driver(config['name'])
        if config['has_setup']:
            args = {a['name']: a['value'] for a in config['setup_arguments']}
            self.driver.setup(**args)

        self.ports = config['ports']
        for port in config['ports']:
            pin = port['name']
            if hasattr(self.driver, 'Pins'):
                pin = self.driver.Pins(pin)
            self.driver.map_pin(port['id'], pin)

        self.inputs = [p for p in config['ports'] if p['type'] & 4]
        ps = [p['id'] for p in self.inputs]
        self.inputs = {
            p['alias']: lambda id=p['id']: self.driver.read(id)
            for p in self.inputs
        }
        cs = [c for c in config['calibrations'] if c['port'] in ps]
        cs = {
            c['alias']:
            lambda p=c['port'], f=c['formula']: self._read_calibrated(p, f)
            for c in cs
        }
        self.inputs = {**self.inputs, **cs}

        self.outputs = [p for p in config['ports'] if p['type'] & (8 | 16)]
        ps = {p['id']: p['type'] for p in self.outputs}
        self.outputs = {p['alias']: lambda x, id=p['id'], t=p['type']: self.driver.write(id, x, (t & 16) != 0)
                        for p in self.outputs}
        cs = [c for c in config['calibrations'] if c['port'] in ps.keys()]
        cs = {c['alias']: lambda x, p=c['port'], f=c['formula'], t=ps[c['port']]: self._write_calibrated(p, f, x, (t & 16) != 0)
              for c in cs}
        self.outputs = {**self.outputs, **cs}

    def _read_calibrated(self, port, formula):
        local = {'x': self.driver.read(port), 'math': math, 'numpy': numpy}
        code = compile('y=%s' % formula, '_string_', 'exec')
        exec(code, local, local)
        return local['y']

    def _write_calibrated(self, port, formula, value, pwm):
        local = {'x': value, 'math': math, 'numpy': numpy}
        exec('y=%s' % formula, local, local)
        self.driver.write(port, local['y'], pwm)

    def read(self, port):
        f = self.inputs.get(port, None)
        if f is None:
            raise Exception(f'Port {port} not configured')
        return f()

    def write(self, port, value):
        f = self.outputs.get(port, None)
        if f is None:
            raise Exception(f'Port {port} not configured')
        f(value)



# -*- coding: utf-8; -*-
#
# Copyright (c) 2016 Álan Crístoffer
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import json
from pathlib import Path

__printed = False


def DatabaseV1():
    global __printed
    try:
        with open('%s/.moirai.json' % str(Path.home())) as f:
            cfgstr = f.read()
            if cfgstr:
                cfg = json.loads(cfgstr)
                db = cfg.get('database', None)
                if db:
                    adapter = db.get('adapter', 'mongodb')
                    host = db.get('host', '127.0.0.1')
                    port = db.get('port', 3306)
                    username = db.get('username', None)
                    password = db.get('password', None)
                    if adapter == 'mongodb':
                        if not __printed:
                            print('Using MongoDB')
                            __printed = True
                        from moirai.database.mongodb import DatabaseV1
                        return DatabaseV1()
                    else:
                        if not __printed:
                            print('Using MySQL')
                            __printed = True
                        from moirai.database.mysql import DatabaseV1
                        return DatabaseV1(host, port, username, password)
    except Exception:
        print('Falling back to MongoDB...')
        from moirai.database.mongodb import DatabaseV1
        return DatabaseV1()



# -*- coding: utf-8; -*-
#
# Copyright (c) 2016 Álan Crístoffer
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from moirai.decorators import decorate_all_methods, dont_raise, log
from moirai.hardware.controller import Controller
from moirai.hardware.free import Free
from moirai.hardware.model_simulation import ModelSimulation
from moirai.hardware.pid import PID
from moirai.hardware.system_response_tests import SystemResponseTest


@decorate_all_methods(dont_raise)
@decorate_all_methods(log)
class CommandProcessor(object):
    """
    Processes commands for this process
    """

    def __init__(self, handler):
        self.handler = handler
        self.pid = PID.instance()
        self.free = Free.instance()

    def process_command(self, sender, cmd, args):
        """
        Handles commands received from other processes.
        """
        cmd = cmd.lower().strip()
        method = getattr(self, cmd, None)
        if method:
            ret = method(args)
            if ret:
                self.handler.send_command(sender, cmd, ret)

    def init(self, _):
        """
        Initialization function for this process.
        """
        pass

    def run_test(self, test):
        test = SystemResponseTest(test)
        test.run()

    def run_controller(self, controller):
        controller = Controller(controller)
        controller.run()

    def run_simulation(self, arg):
        data, pipe = arg
        sim = ModelSimulation(data)
        result = sim.run()
        pipe.send(result)

    def run_pid(self, args):
        self.handler.flush()
        self.pid.run(args)

    def run_free(self, args):
        self.handler.flush()
        self.free.run(args)



# -*- coding: utf-8; -*-
#
# Copyright (c) 2016 Álan Crístoffer
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import ahio
import io
import zipfile
import hashlib
import json
import dateutil.parser
import logging
import os.path
import tempfile
import scipy.io as sio
from multiprocessing import Pipe

from bson import json_util
from enum import Enum

from cheroot.wsgi import Server
from cheroot.wsgi import PathInfoDispatcher
from flask import Flask, request, send_file
from werkzeug.serving import WSGIRequestHandler
from moirai.database import DatabaseV1
from moirai.hardware import Hardware
from moirai import __version__


class APIv1:
    """
    Starts a WebServer for the API endpoint.
    """

    def __init__(self, processHandler):
        self.app = Flask(__name__)
        self.database = DatabaseV1()
        self.hardware = Hardware()
        self.ph = processHandler
        logging.getLogger('werkzeug').setLevel(logging.ERROR)

    def run(self):
        """
        Entry point for the class. Adds routes and starts listening.
        """

        @self.app.after_request
        def add_header(response):
            response.headers['Cache-Control'] = 'no-store'
            return response

        self.app.add_url_rule('/', view_func=lambda: 'Moirai Control System\n')
        self.app.add_url_rule('/login', view_func=self.login, methods=['POST'])
        self.app.add_url_rule(
            '/version', view_func=self.version, methods=['GET'])
        self.app.add_url_rule(
            '/set-password', view_func=self.set_password, methods=['POST'])
        self.app.add_url_rule(
            '/last_error', view_func=self.last_error, methods=['GET'])
        self.app.add_url_rule(
            '/hardware/drivers',
            view_func=self.hardware_drivers,
            methods=['GET'])
        self.app.add_url_rule(
            '/hardware/configuration',
            view_func=self.hardware_set_configuration,
            methods=['POST'])
        self.app.add_url_rule(
            '/hardware/configuration',
            view_func=self.hardware_get_configuration,
            methods=['GET'])
        self.app.add_url_rule(
            '/system_response/tests',
            view_func=self.system_response_get_tests,
            methods=['GET'])
        self.app.add_url_rule(
            '/system_response/tests',
            view_func=self.system_response_set_tests,
            methods=['POST'])
        self.app.add_url_rule(
            '/system_response/test/run',
            view_func=self.system_response_run,
            methods=['POST'])
        self.app.add_url_rule(
            '/system_response/test/stop',
            view_func=self.system_response_stop,
            methods=['GET'])
        self.app.add_url_rule(
            '/live_graph/tests',
            view_func=self.live_graph_list_tests,
            methods=['GET'])
        self.app.add_url_rule(
            '/live_graph/test',
            view_func=self.live_graph_get_test,
            methods=['POST'])
        self.app.add_url_rule(
            '/live_graph/test/remove',
            view_func=self.live_graph_remove_test,
            methods=['POST'])
        self.app.add_url_rule(
            '/live_graph/test/export',
            view_func=self.live_graph_export_mat,
            methods=['POST'])
        self.app.add_url_rule(
            '/controllers', view_func=self.controller_set, methods=['POST'])
        self.app.add_url_rule(
            '/controllers', view_func=self.controller_get, methods=['GET'])
        self.app.add_url_rule(
            '/controllers/run',
            view_func=self.controller_run,
            methods=['POST'])
        self.app.add_url_rule(
            '/controllers/stop',
            view_func=self.controller_stop,
            methods=['GET'])
        self.app.add_url_rule(
            '/db/dump', view_func=self.dump_database, methods=['GET'])
        self.app.add_url_rule(
            '/db/restore', view_func=self.restore_database, methods=['POST'])
        self.app.add_url_rule(
            '/simulation/run',
            view_func=self.model_simulation_run,
            methods=['POST'])
        self.app.add_url_rule(
            '/pid/run', view_func=self.pid_run, methods=['POST'])
        self.app.add_url_rule(
            '/free/run', view_func=self.free_run, methods=['POST'])

        d = PathInfoDispatcher({'/': self.app})
        self.server = Server(('0.0.0.0', 5000), d)
        self.server.start()

    def stop(self):
        self.server.stop()

    def version(self):
        return json.dumps({'version': __version__})

    def verify_token(self):
        """
        Verifies the token sent as a HTTP Authorization header.
        """
        try:
            authorization = request.headers.get('Authorization')
            token = authorization.split(' ')[-1]
            return self.database.verify_token(token)
        except Exception as e:
            return False

    def login(self):
        """
        Authenticates the user.
        Should be called with a POST request containing the following body:

        {
            "password": string
        }

        @returns:
            On success, HTTP 200 Ok and body:

            {
                "token": string
            }

            On failure, HTTP 403 Unauthorized and body:

            {}
        """
        password = request.json.get('password', '')
        hasher = hashlib.sha512()
        hasher.update(bytes(password, 'utf-8'))
        password = hasher.hexdigest()
        saved_password = self.database.get_setting('password')
        if saved_password == password:
            return json.dumps({'token': self.database.generate_token()})
        return '{}', 403

    def set_password(self):
        """
        Sets the password. Required body:

        {
            "password": string
        }

        @returns:
            On success, HTTP 200 Ok and body:

            {}

            On failure, HTTP 403 Unauthorized and body:

            {}
        """
        if not self.verify_token():
            return '{}', 403
        password = request.json.get('password', '')
        hasher = hashlib.sha512()
        hasher.update(bytes(password, 'utf-8'))
        password = hasher.hexdigest()
        self.database.set_setting('password', password)
        return '{}'

    def last_error(self):
        """
        Returns the last error in the database.

        @returns:
            {
                message: string
            }
        """
        if not self.verify_token():
            return '{}', 403
        error = self.database.get_setting('test_error')
        return json.dumps({'message': error})

    def hardware_drivers(self):
        """
        Returns a JSON object with all the available drivers, their setups, if
        any, and ports, if listable.

        @returns:
            {
                name: string,
                has_setup: boolean,
                setup_arguments: [
                    {
                        name: string,
                        default_value: string
                    }
                ],
                ports: [
                    {
                        id: number,
                        name: string,
                        analog: {
                            input: boolean,
                            output: boolean,
                            read_range: [number, number],
                            write_range: [number, number]
                        }
                        digital: {
                            input: boolean,
                            output: boolean,
                            pwm: boolean
                        }
                    }
                ]
            }
        """
        if not self.verify_token():
            return '{}', 403
        drivers = self.hardware.list_drivers()
        drivers = [{
            'name':
            driver,
            'has_setup':
            self.hardware.driver_has_setup(driver),
            'setup_arguments':
            self.hardware.driver_setup_arguments(driver),
            'ports':
            self.__ports_for_driver(driver)
        } for driver in drivers]
        return json.dumps(drivers)

    def hardware_set_configuration(self):
        """
        Saves the given driver configuration. It must be a POST request with
        the following body:

        {
            name: string,
            setup_arguments: [
                {
                    name: string,
                    value: string
                }
            ],
            ports: [
                {
                    id: number,
                    name: string | number,
                    alias: string,
                    type: number,
                    defaultValue: string
                }
            ],
            configurations: [
                {
                    port: number,
                    alias: string,
                    formula: string
                }
            ]
        }

        @returns:
            On success, HTTP 200 Ok and body:

            {}

            On failure, HTTP 403 Unauthorized and body:

            {}
        """
        if not self.verify_token():
            return '{}', 403

        configuration = request.json
        self.database.set_setting('hardware_configuration', configuration)
        return '{}'

    def hardware_get_configuration(self):
        """
        Returns the saved driver configuration. It must be a GET request.

        @returns:
            On success, HTTP 200 Ok and body:

            {
                name: string,
                setup_arguments: [
                    {
                        name: string,
                        value: string
                    }
                ],
                ports: [
                    {
                        id: number,
                        name: string | number,
                        alias: string,
                        type: number,
                        defaultValue: string
                    }
                ],
                configurations: [
                    {
                        port: number,
                        alias: string,
                        formula: string
                    }
                ]
            }

            or

            {} if there is no configuration saved

            On failure, HTTP 403 Unauthorized and body:

            {}
        """
        if not self.verify_token():
            return '{}', 403

        config = self.database.get_setting('hardware_configuration') or {}
        return json.dumps(config)

    def system_response_get_tests(self):
        """
        Returns the saved system response tests. It must be a GET request.

        @returns:
            On success, HTTP 200 Ok and body:

            [{
                id: number
                name: string
                type: string
                inputs: string[]
                output: string[]
                points: [{
                    x: number
                    y: number
                }]
                fixedOutputs: [{
                    alias: string
                    value: number
                }]
                logRate: number
            }]

            or

            [] if there is no configuration saved

            On failure, HTTP 403 Unauthorized and body:

            []
        """
        if not self.verify_token():
            return '{}', 403

        tests = self.database.get_setting('system_response_tests') or []
        return json.dumps(tests)

    def system_response_set_tests(self):
        """
        Sets the saved system response tests. It must be a POST request with
        the following body:

        [{
            id: number
            name: string
            type: string
            inputs: string[]
            output: string[]
            points: [{
                x: number
                y: number
            }]
            fixedOutputs: [{
                alias: string
                value: number
            }]
            logRate: number
        }]

        @returns:
            On success, HTTP 200 Ok and body:

            {}

            On failure, HTTP 403 Unauthorized and body:

            {}
        """
        if not self.verify_token():
            return '{}', 403

        tests = request.json
        self.database.set_setting('system_response_tests', tests)
        return '{}'

    def system_response_run(self):
        """
        Runs the given test. It must be a POST request with the following body:

        {
            test: number
        }

        @returns:
            On success, HTTP 200 Ok and body:

            {}

            On failure, HTTP 403 Unauthorized and body:

            {}
        """
        if not self.verify_token():
            return '{}', 403

        test = request.json['test']
        self.ph.send_command("hardware", "run_test", test)
        return '{}'

    def system_response_stop(self):
        """
        Stops the running test. It must be a GET request.

        @returns:
            On success, HTTP 200 Ok and body:

            {}

            On failure, HTTP 403 Unauthorized and body:

            {}
        """
        if not self.verify_token():
            return '{}', 403

        self.database.set_setting('current_test', None)

        return '{}'

    def live_graph_list_tests(self):
        """
        Returns a list of available graphs. It must be a GET request.

        @returns:
            On success, HTTP 200 Ok and body:

            [
                {
                    name: string
                    date: string (ISO 8601)
                    running: boolean
                }
            ]

            On failure, HTTP 403 Unauthorized and body:

            []
        """
        if not self.verify_token():
            return '{}', 403
        running_test = self.database.get_setting('current_test')
        tests = self.database.list_test_data()
        if len(tests) == 0:
            return '[]'

        last_date = max([test['date'] for test in tests])
        tests = [{
            'name':
            t['name'],
            'date':
            t['date'].isoformat(),
            'running':
            t['name'] == running_test and t['date'] == last_date
        } for t in tests]

        return json.dumps(tests)

    def live_graph_get_test(self):
        """
        Returns a graph. It must be a POST request with following body:

        {
            test: string
            start_time: string (ISO 8601)
            skip?: number
        }

        @returns:
            On success, HTTP 200 Ok and body:

            [
                {
                    sensor: string
                    time: string | number
                    value: string | number
                }
            ]

            On failure, HTTP 403 Unauthorized and body:

            []
        """
        if not self.verify_token():
            return '{}', 403

        test = request.json['test']
        start_time = dateutil.parser.parse(request.json['start_time'])
        skip = request.json.get('skip', 0)

        points = self.database.get_test_data(test, start_time, skip)

        return json.dumps(points)

    def live_graph_remove_test(self):
        """
        Deletes a test. It must be a POST request with following body:

        {
            test: string
            start_time: string (ISO 8601)
        }

        or a list of elements like that.

        @returns:
            On success, HTTP 200 Ok and body:

            []

            On failure, HTTP 403 Unauthorized and body:

            []
        """
        if not self.verify_token():
            return '{}', 403

        ts = request.json
        if isinstance(ts, list):
            ts = [{
                'test': t['test'],
                'start_time': dateutil.parser.parse(t['start_time'])
            } for t in ts]
        else:
            ts = {
                'test': ts['test'],
                'start_time': dateutil.parser.parse(ts['start_time'])
            }

        self.database.remove_test(ts)

        return '[]'

    def live_graph_export_mat(self):
        """
        Generates a MATLAB's MAT file from data. It must be a POST request with
        following body:

        {
            test: string
            start_time: string (ISO 8601)
        }

        @returns:
            On success, HTTP 200 Ok and body:

            file-contents

            On failure, HTTP 403 Unauthorized and body:

            {}
        """
        if not self.verify_token():
            return '{}', 403

        test = request.json['test']
        start_time = dateutil.parser.parse(request.json['start_time'])
        v = request.json['variables']

        ks = list(v.keys())
        ds = self.database.get_filtered_test_data(test, start_time, ks)
        ts = ds[0]['time']
        ds = {v.get(d['sensor'], d['sensor']): d['values'] for d in ds}
        ds['t'] = ts

        directory = tempfile.gettempdir()
        file_path = os.path.join(directory, 'test.mat')

        sio.savemat(file_path, ds)

        f = open(file_path, 'rb')

        return send_file(f, as_attachment=True, attachment_filename='data.mat')

    def controller_set(self):
        """
        Saves a controller. It must be a POST request with following body:

        [{
            id: number
            name: string
            tau: number
            runTime: number
            before: string
            controller: string
            after: string
            inputs: string[]
        }]

        @returns:
            On success, HTTP 200 Ok and body:

            {}

            On failure, HTTP 403 Unauthorized and body:

            {}
        """
        if not self.verify_token():
            return '{}', 403

        cs = request.json
        self.database.set_setting('controllers', cs)
        return '{}'

    def controller_get(self):
        """
        Get saved controllers. It must be a GET request.

        @returns:
            On success, HTTP 200 Ok and body:

            [{
                id: number
                name: string
                tau: number
                runTime: number
                before: string
                controller: string
                after: string
                inputs: string[]
            }]

            On failure, HTTP 403 Unauthorized and body:

            {}
        """
        if not self.verify_token():
            return '{}', 403

        cs = self.database.get_setting('controllers') or []
        return json.dumps(cs)

    def controller_run(self):
        """
        Runs the given controller. It must be a POST request with the following
        body:

        {
            controller: number
        }

        @returns: On success, HTTP 200 Ok and body:

            {}

            On failure, HTTP 403 Unauthorized and body:

            {}
        """
        if not self.verify_token():
            return '{}', 403

        controller = request.json['controller']
        self.ph.send_command("hardware", "run_controller", controller)
        return '{}'

    def controller_stop(self):
        """
        Stops the running controller. It must be a GET request.

        @returns:
            On success, HTTP 200 Ok and body:

            {}

            On failure, HTTP 403 Unauthorized and body:

            {}
        """
        if not self.verify_token():
            return '{}', 403

        self.database.set_setting('current_test', None)

        return '{}'

    def dump_database(self):
        """
        Dumps the database collections. Use for backup.

        @returns:
            On success, HTTP 200 Ok and body:

            {
                settings: []
                test_sensor_values: []
            }

            On failure, HTTP 403 Unauthorized and body:

            {}
        """
        if not self.verify_token():
            return '{}', 403

        settings, test_sensor_values = self.database.dump_database()
        data = {'settings': settings, 'test_sensor_values': test_sensor_values}
        jsondata = json.dumps(data, default=json_util.default)

        del data, settings, test_sensor_values  # release memory

        zbuffer = io.BytesIO()
        with zipfile.ZipFile(zbuffer, "a", zipfile.ZIP_LZMA) as zip_file:
            zip_file.writestr('dump', jsondata)
        zbuffer.seek(0, 0)
        return send_file(
            zbuffer,
            as_attachment=True,
            attachment_filename='dump.zip',
            mimetype="application/octet-stream")

    def restore_database(self):
        """
        Restore the database collections. It must be a POST request with following body:

        [{
            settings: []
            test_sensor_values: []
        }]

        @returns:
            On success, HTTP 200 Ok and body:

            {}

            On failure, HTTP 403 Unauthorized and body:

            {}
        """
        if not self.verify_token():
            return '{}', 403
        db = self.database
        file = tempfile.TemporaryFile()
        file.write(request.files['file'].read())
        file.seek(0)
        with zipfile.ZipFile(file, "r", zipfile.ZIP_LZMA) as zip_file:
            jsondata = zip_file.read('dump')
            d = json.loads(jsondata, object_hook=json_util.object_hook)
            del jsondata
            db.restore_database(d['settings'], d['test_sensor_values'])
        return '{}'

    def __ports_for_driver(self, driver):
        """
        Returns a list of encoded ports for the given driver.
        """
        ps = self.hardware.driver_ports(driver)
        ps = [self.__encode_port(p) for p in ps]
        return ps

    def __encode_port(self, port):
        """
        If the port is a Enum, use its value. If it's a value that can be dumped
        to JSON, use it, if not, raise an exception.
        """
        port_id = port['id']
        if isinstance(port['id'], Enum):
            port_id = port['id'].value
        else:
            try:
                port_id = json.dumps(port['id'])
            except:
                pass
        port['id'] = port_id
        return port

    def __decode_port(self, driver, port):
        """
        If the ID in port should be converted to Enum, do it. Else, keep it as
        it is.
        """
        if driver not in ahio.list_available_drivers():
            return port
        driver = ahio.new_driver(driver)
        ps = driver.available_pins()
        if len(ps) > 0:
            p = ps[0]
            if isinstance(p['id'], Enum):
                port['id'] = p['id'].__class__(port['id'])
        return port

    def model_simulation_run(self):
        """
        Simulates the model. It must be a POST request with the following body:

        {
            model: string
            x0: string
            u: string
            duration: string
        }

        @returns: On success, HTTP 200 Ok and body:

            {}

            On failure, HTTP 403 Unauthorized and body:

            {}
        """
        if not self.verify_token():
            return '{}', 403

        args = request.json
        p1, p2 = Pipe()
        self.ph.send_command("hardware", "run_simulation", (args, p1))
        return p2.recv()

    def pid_run(self):
        """
        Runs the PID controller for controlled-variable y (port name expected)
        and control-variable u (port name expected).

        {
            Kp: number
            Ki: number
            Kd: number
            dt: number
            u: string
            y: string
            r: string
            umax: number
            umin: number
            fixedOutputs: {
                            alias: string
                            value: number
                          }[]
        }

        @returns: On success, HTTP 200 Ok and body:

            {}

            On failure, HTTP 403 Unauthorized and body:

            {}
        """
        if not self.verify_token():
            return '{}', 403

        args = request.json
        self.ph.send_command("hardware", "run_pid", args)
        return '{}'

    def free_run(self):
        """
        Controls the plant outputs freely.

        {
            dt: number
            inputs: string[]
            outputs: {
                        alias: string
                        value: number
                     }[]
        }

        @returns: On success, HTTP 200 Ok and body:

            {}

            On failure, HTTP 403 Unauthorized and body:

            {}
        """
        if not self.verify_token():
            return '{}', 403

        args = request.json
        self.ph.send_command("hardware", "run_free", args)
        return '{}'


# -*- coding: utf-8; -*-
#
# Copyright (c) 2016 Álan Crístoffer
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import datetime
import math
import sys
import time
import traceback

from moirai.database import DatabaseV1
from moirai.hardware.configured_hardware import ConfiguredHardware
from moirai.hardware.timer import Timer


class PID(object):
    __instance = None

    @classmethod
    def instance(cls):
        if not PID.__instance:
            PID.__instance = PID()
        return PID.__instance

    def __init__(self):
        self.last_run = time.time()
        self.running = False
        self.timer = Timer(math.inf, 1)
        self.Kp = 0
        self.Ki = 0
        self.Kd = 0
        self.y = ''
        self.u = ''
        self.le = 0
        self.se = 0
        self.li = 0
        self.ls = 0
        self.fixedOutputs = []
        self.hardware = None
        self.db = DatabaseV1()
        self.start_time = datetime.datetime.utcnow()
        self.locks = []

    def is_valid(self):
        return (time.time() - self.last_run) < max(2 * self.timer.interval, 1)

    def run(self, data=None):
        try:
            if data:
                self.last_run = time.time()
                self.running = True
                self.timer.interval = float(data['dt'])
                self.Kp = float(data['Kp'])
                self.Ki = float(data['Ki'])
                self.Kd = float(data['Kd'])
                self.r = float(data['r'])
                self.li = float(data['umin'])
                self.ls = float(data['umax'])
                self.y = data['y']
                self.u = data['u']
                self.fixedOutputs = data['fixedOutputs']
                config = self.db.get_setting('hardware_configuration')
                self.locks = [self.interlock(l) for l in config['interlocks']]

            if not self.is_valid():
                self.running = False

            if self.running:
                if not self.hardware:
                    self.db.set_setting('test_error', '')
                    self.timer = Timer(math.inf, float(data['dt']))
                    self.start_time = datetime.datetime.utcnow()
                    self.hardware = ConfiguredHardware()
                    self.last_run = time.time()

                    for output in self.fixedOutputs:
                        self.hardware.write(output['alias'], output['value'])

                    self.db.save_test_sensor_value('PID', self.y, 0, 0,
                                                   self.start_time)
                    self.db.save_test_sensor_value('PID', self.u, 0, 0,
                                                   self.start_time)
                    self.db.save_test_sensor_value('PID', 'R', self.r, 0,
                                                   self.start_time)

                self.timer.sleep()

                y = self.hardware.read(self.y)
                e = self.r - y
                de = e - self.le
                self.se += e
                u = self.Kp * e + self.Kd * de + self.Ki * self.se
                u = sorted([self.li, self.ls, u])[1]
                self.hardware.write(self.u, u)
                self.le = e

                self.db.save_test_sensor_value('PID', self.y, y,
                                               self.timer.elapsed(),
                                               self.start_time)
                self.db.save_test_sensor_value('PID', self.u, u,
                                               self.timer.elapsed(),
                                               self.start_time)
                self.db.save_test_sensor_value('PID', 'R', self.r,
                                               self.timer.elapsed(),
                                               self.start_time)

                for lock in self.locks:
                    lock()
            elif self.hardware:
                self.shutdown()
        except Exception as e:
            print(e)
            self.running = False
            self.shutdown()

    def shutdown(self):
        if self.hardware:
            self.off_values = {
                p['alias']: float(p['defaultValue'])
                for p in self.hardware.ports if p['type'] & (8 | 16)
            }
            for k, v in self.off_values.items():
                self.hardware.write(k, v)
            self.hardware = None

    def interlock(self, lock):
        try:
            code = compile('y=%s' % lock['expression'], '_string_', 'exec')
        except SyntaxError as err:
            error = err.__class__.__name__
            detail = err.args[0]
            line = err.lineno
            error_string = '%s on %s:%s: %s' % (error, 'pid', line, detail)
            print(error_string)
            self.db.set_setting('test_error', error_string)
            self.running = False
            return

        def f(code=code, lock=lock):
            try:
                value = self.hardware.read(lock['sensor'])
                scope = {'x': value}
                exec(code, None, scope)
            except Exception as err:
                error = err.__class__.__name__
                detail = err.args[0]
                cl, exc, tb = sys.exc_info()
                tb = self.stringify_tb(traceback.extract_tb(tb))
                error_string = '%s: %s\n%s' % (error, detail, tb)
                print(error_string)
                self.db.set_setting('test_error', error_string)
                raise Exception('Interlock')
            if scope['y']:
                self.hardware.write(lock['actuator'], lock['actuatorValue'])
                self.db.set_setting('test_error', 'Interlock')
                raise Exception('Interlock')

        return f

    def stringify_tb(self, tb):
        msg = 'Traceback:\n\t'
        msg += '\n\t'.join(
            ['%s:%s in %s' % (t.filename, t.lineno, t.name) for t in tb])
        return msg



#!/usr/bin/env python3
# -*- coding: utf-8; -*-
#
# Copyright (c) 2016 Álan Crístoffer
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import codecs
import glob
import shutil
from os import path

import moirai
from setuptools import Command, find_packages, setup

PWD = path.abspath(path.dirname(__file__))
with codecs.open(path.join(PWD, 'README.md'), encoding='utf-8') as f:
    LONG_DESCRIPTION = f.read()


class CleanCommand(Command):
    """Custom clean command to tidy up the project root."""
    CLEAN_FILES = [
        './build', './dist', './*.pyc', './*.tgz', './*.egg-info',
        './**/__pycache__'
    ]

    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        cwd = path.realpath(path.curdir)
        ps = [glob.glob(p) for p in self.CLEAN_FILES]
        ps = [path.normpath(path.join(cwd, p)) for p in sum(ps, [])]
        ps = [p for p in ps if p.startswith(cwd)]
        for p in ps:
            print('removing %s' % p)
            shutil.rmtree(p)


setup(
    name='moirai',
    version=moirai.__version__,
    description='Digital Control Manager Backend',
    long_description=LONG_DESCRIPTION,
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English', 'Operating System :: OS Independent',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Scientific/Engineering :: Human Machine Interfaces'
    ],
    keywords='digital control manager',
    author='Álan Crístoffer',
    author_email='acristoffers@gmail.com',
    url='https://www.github.com/acristoffers/moirai',
    packages=find_packages(),
    license="MIT",
    install_requires=[
        'appdirs', 'ahio', 'Flask', 'pymongo', 'python-dateutil', 'numpy',
        'scipy', 'cheroot'
    ],
    entry_points={'console_scripts': ['moirai = moirai.moirai:start']},
    cmdclass={'clean': CleanCommand})



# -*- coding: utf-8; -*-
#
# Copyright (c) 2016 Álan Crístoffer
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import ahio
import datetime

import numpy as np
import math
import threading
import time

import os
import sys
import traceback

from moirai.database import DatabaseV1
from moirai.hardware.configured_hardware import ConfiguredHardware
from moirai.hardware.timer import Timer, Finished


class Controller(object):
    def __init__(self, controller_id):
        self.db = DatabaseV1()
        cs = self.db.get_setting('controllers')
        self.cs = next((c for c in cs if c['id'] == controller_id), None)
        if self.cs is None:
            raise Exception('Controller not found')
        self.hardware = ConfiguredHardware()
        configuration = self.db.get_setting('hardware_configuration')
        self.locks = [self.interlock(l) for l in configuration['interlocks']]
        self.off_values = {
            p['alias']: float(p['defaultValue'])
            for p in configuration['ports'] if p['type'] & (8 | 16)
        }
        self.running = True
        self.lock = threading.Lock()
        self.after = None

    def interlock(self, lock):
        try:
            code = compile('y=%s' % lock['expression'], '_string_', 'exec')
        except SyntaxError as err:
            error = err.__class__.__name__
            detail = err.args[0]
            line = err.lineno
            error_string = '%s on %s:%s: %s' % (error, scope, line, detail)
            print(error_string)
            self.db.set_setting('test_error', error_string)
            self.running = False
            return

        def f(code=code, lock=lock):
            try:
                value = self.hardware.read(lock['sensor'])
                scope = {'x': value}
                exec(code, None, scope)
            except Exception as err:
                error = err.__class__.__name__
                detail = err.args[0]
                cl, exc, tb = sys.exc_info()
                tb = self.stringify_tb(traceback.extract_tb(tb))
                error_string = '%s: %s\n%s' % (error, detail, tb)
                print(error_string)
                self.db.set_setting('test_error', error_string)
                raise Exception('Interlock')
            if scope['y']:
                self.hardware.write(lock['actuator'], lock['actuatorValue'])
                self.db.set_setting('test_error', 'Interlock')
                raise Exception('Interlock')

        return f

    def lock_forever(self):
        interval = float(self.cs['tau'])
        while self.running:
            time.sleep(interval)
            self.lock.acquire()
            try:
                for lock in self.locks:
                    lock()
            except:
                try:
                    self.running = False
                    if self.after is not None:
                        ins = self.cs['inputs']
                        inputs = {s: self.hardware.read(s) for s in ins}
                        plocals = {
                            'inputs': inputs,
                            'outputs': dict(),
                            'np': np,
                            'math': math
                        }

                        exec(self.after, plocals, plocals)

                        for actuator, value in plocals['outputs'].items():
                            self.hardware.write(actuator, value)
                except Exception as err:
                    error = err.__class__.__name__
                    detail = err.args[0]
                    cl, exc, tb = sys.exc_info()
                    tb = self.stringify_tb(traceback.extract_tb(tb))
                    error_string = '%s: %s\n%s' % (error, detail, tb)
                    print(error_string)
                    self.db.set_setting('test_error', error_string)
            self.lock.release()

    def run(self):
        self.db.set_setting('current_test', self.cs['name'])
        self.db.set_setting('test_error', None)

        after = None
        self.running = True

        try:
            scope = 'before'
            before = compile(self.cs['before'], 'before', 'exec')
            scope = 'controller'
            controller = compile(self.cs['controller'], 'controller', 'exec')
            scope = 'after'
            after = compile(self.cs['after'], 'after', 'exec')
            self.after = after

            thread = threading.Thread(target=self.lock_forever)
            thread.start()
            thread.isDaemon = True

            inputs = {s: self.hardware.read(s) for s in self.cs['inputs']}
            plocals = {
                'inputs': inputs,
                'outputs': dict(),
                's': dict(),
                'dt': float(self.cs['tau']),
                'np': np,
                'math': math
            }

            for k, v in self.off_values.items():
                self.hardware.write(k, v)

            exec(before, plocals, plocals)

            state = plocals['s']

            for k, v in plocals['outputs'].items():
                self.hardware.write(k, v)

            run_time = int(self.cs['runTime'])
            interval = float(self.cs['tau'])
            t = Timer(run_time, interval)
            start_time = datetime.datetime.utcnow()
            time = 0

            while self.db.get_setting(
                    'current_test') is not None and self.running:
                self.lock.acquire()
                inputs = {s: self.hardware.read(s) for s in self.cs['inputs']}
                self.lock.release()

                plocals = {
                    'inputs': inputs,
                    'outputs': dict(),
                    's': state,
                    'log': dict(),
                    't': time,
                    'dt': interval,
                    'np': np,
                    'math': math
                }

                exec(controller, plocals, plocals)

                plocals['log'] = {
                    **plocals['log'],
                    **plocals['inputs'],
                    **plocals['outputs']
                }

                self.lock.acquire()
                for k, v in plocals['outputs'].items():
                    self.hardware.write(k, v)

                for k, v in plocals['log'].items():
                    cid = self.cs['name']
                    self.db.save_test_sensor_value(cid, k, v, time, start_time)
                self.lock.release()

                state = plocals['s']

                t.sleep()
                time = t.elapsed()
        except Finished:
            pass
        except SyntaxError as err:
            error = err.__class__.__name__
            detail = err.args[0]
            line = err.lineno
            error_string = '%s on %s:%s: %s' % (error, scope, line, detail)
            print(error_string)
            self.db.set_setting('test_error', error_string)
        except Exception as err:
            error = err.__class__.__name__
            detail = err.args[0]
            cl, exc, tb = sys.exc_info()
            tb = self.stringify_tb(traceback.extract_tb(tb))
            error_string = '%s: %s\n%s' % (error, detail, tb)
            print(error_string)
            self.db.set_setting('test_error', error_string)

        try:
            if after is not None:
                inputs = {s: self.hardware.read(s) for s in self.cs['inputs']}
                plocals = {
                    'inputs': inputs,
                    'outputs': dict(),
                    'np': np,
                    'math': math
                }

                exec(after, plocals, plocals)

                for actuator, value in plocals['outputs'].items():
                    self.hardware.write(actuator, value)

            for k, v in self.off_values.items():
                self.hardware.write(k, v)
        except Exception as err:
            error = err.__class__.__name__
            detail = err.args[0]
            cl, exc, tb = sys.exc_info()
            tb = self.stringify_tb(traceback.extract_tb(tb))
            error_string = '%s: %s\n%s' % (error, detail, tb)
            print(error_string)
            self.db.set_setting('test_error', error_string)

        self.db.set_setting('current_test', None)
        self.db.close()
        self.running = False

    def stringify_tb(self, tb):
        msg = 'Traceback:\n\t'
        msg += '\n\t'.join(
            ['%s:%s in %s' % (t.filename, t.lineno, t.name) for t in tb])
        return msg
